/***************************************************************************
 *            geo_photos.c
 *
 *  Sat Apr 12 16:52:15 2008
 *  Copyright  2008  Marcus Bauer
 *  marcus.bauer@gmail.com
 ****************************************************************************/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "globals.h"
#include "geo_photos.h"
#include "converter.h"
#include "util.h"

#include <glib.h>
#include <glib/gprintf.h>
#include <stdio.h>
#include <sqlite3.h>
#include <stdlib.h>

#define PHOTO_DB "geophoto.db"



/* generic prototype, good for testing */
/*
static int
sql_cb(void *unused, int colc, char **colv, char **col_names)
{
	int i;
	for(i=0; i<colc; i++)
	{
		printf("%s = %s\n", col_names[i], colv[i] ? colv[i] : "NULL");
	}

	printf("\n");
	return 0;
}
*/


static int
sql_cb__photo(void *unused, int colc, char **colv, char **col_names)
{
	photo_t *photo = g_new0(photo_t,1);

	printf("*** %s(): \n",__PRETTY_FUNCTION__);

	photo->filename = g_strdup(colv[0]);
	photo->name	= g_strdup(colv[1]);
	photo->lat	= atof(colv[2]);
	photo->lon	= atof(colv[3]);
	photo->desc	= g_strdup(colv[4]);
	
	photo_list = g_slist_prepend(photo_list, photo);

	return 0;
}







void
paint_photos()
{
	//TODO this function partially duplicate code and should be generalised
	
	int pixel_x, pixel_y, x, y;
	float lat, lon;
	GSList *list;
	GdkColor color;
	GError	*error = NULL;
	static GdkPixbuf *photo_icon = NULL;
	static GdkGC *gc;
	
	printf("*** %s(): \n",__PRETTY_FUNCTION__);

	if (!gc)
		gc = gdk_gc_new(pixmap); //TODO unref/non static?
	color.green = 0;
	color.blue = 60000;
	color.red = 0;
	gdk_gc_set_rgb_fg_color(gc, &color);
	

	if(!photo_icon)
	{
		photo_icon = gdk_pixbuf_new_from_file_at_size (
			PACKAGE_PIXMAPS_DIR "/tangogps-photo.png", 24,24,
			&error);
	}

		
	for(list = photo_list; list != NULL; list = list->next)
	{
		photo_t *p = list->data;
	
		lat = deg2rad(p->lat);
		lon = deg2rad(p->lon);
		
		
		// pixel_x,y, offsets
		pixel_x = lon2pixel(global_zoom, lon);
		pixel_y = lat2pixel(global_zoom, lat);
		
		x = pixel_x - global_x;
		y = pixel_y - global_y;
		
		p->screen_x = x;
		p->screen_y = y;
		
		//this should go into another list...
		
		if(!photo_icon)
		{
			gdk_draw_arc (
				pixmap,
				//map_drawable->style->black_gc,
				gc,
				TRUE,			//filled
				x-4, y-4,		// x,y
				8,8,			// width, height
				0,23040);		// start-end angle 64th, from 3h, anti clockwise
		}
		else
		{
			gdk_draw_pixbuf (
				pixmap,
				NULL,
				photo_icon,
				0,0,
				x-12,y-12,
				24,24,
				GDK_RGB_DITHER_NONE, 0, 0);
			
		}
		
		gtk_widget_queue_draw_area (
				map_drawable, 
				x-12, y-12,
				24,24);
		
		printf("PHOTO: %s lat %f - lon %f\n",p->name,p->lat, p->lon);
	}	
}


void
get_photos()
{
	char *sql, *db;
	
	db = g_strconcat(tangogps_dir,"/", PHOTO_DB, NULL);
	
	//TODO: correctly free photo list //g_slist_free(photo_list);
	photo_list = NULL;
	
	//get bbox
	//get lat,lon,file,name,desc  LIMIT 1000
	sql = g_strdup("SELECT * FROM photo LIMIT 1000;");

	//loop: from db into GSList
	sql_execute(db, sql, sql_cb__photo);	
	
	paint_photos();
	
}
